import unittest
from datetime import date
from enum import Enum

from pypika import (
    SYSTEM_TIME,
    AliasedQuery,
    Case,
    ClickHouseQuery,
    EmptyCriterion,
    Index,
    MSSQLQuery,
    MySQLQuery,
    NullValue,
    OracleQuery,
    Order,
    PostgreSQLQuery,
    Query,
    QueryException,
    RedshiftQuery,
    SQLLiteQuery,
    Table,
    Tables,
    VerticaQuery,
)
from pypika import (
    Field as F,
)
from pypika import (
    functions as fn,
)
from pypika.terms import ValueWrapper

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class SelectTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_empty_query(self):
        q = Query.from_("abc")

        self.assertEqual("", str(q))

    def test_select_no_from(self):
        q = Query.select(1)

        self.assertEqual("SELECT 1", str(q))

    def test_select_no_with_alias_from(self):
        q = Query.select(ValueWrapper(1, "test"))

        self.assertEqual('SELECT 1 "test"', str(q))

    def test_select_no_from_with_field_raises_exception(self):
        with self.assertRaises(QueryException):
            Query.select("asdf")

    def test_select__star(self):
        q = Query.from_("abc").select("*")

        self.assertEqual('SELECT * FROM "abc"', str(q))

    def test_select__table_schema(self):
        q = Query.from_(Table("abc", "schema1")).select("*")

        self.assertEqual('SELECT * FROM "schema1"."abc"', str(q))

    def test_select__table_schema_with_multiple_levels_as_tuple(self):
        q = Query.from_(Table("abc", ("schema1", "schema2"))).select("*")

        self.assertEqual('SELECT * FROM "schema1"."schema2"."abc"', str(q))

    def test_select__table_schema_with_multiple_levels_as_list(self):
        q = Query.from_(Table("abc", ["schema1", "schema2"])).select("*")

        self.assertEqual('SELECT * FROM "schema1"."schema2"."abc"', str(q))

    def test_select__star__replacement(self):
        q = Query.from_("abc").select("foo").select("*")

        self.assertEqual('SELECT * FROM "abc"', str(q))

    def test_select__distinct__single(self):
        q = Query.from_("abc").select("foo").distinct()

        self.assertEqual('SELECT DISTINCT "foo" FROM "abc"', str(q))

    def test_select__distinct__multi(self):
        q = Query.from_("abc").select("foo", "bar").distinct()

        self.assertEqual('SELECT DISTINCT "foo","bar" FROM "abc"', str(q))

    def test_select__column__single__str(self):
        q = Query.from_("abc").select("foo")

        self.assertEqual('SELECT "foo" FROM "abc"', str(q))

    def test_select__column__single__alias__str(self):
        q = Query.from_(self.table_abc).select(self.table_abc.foo.as_("bar"))

        self.assertEqual('SELECT "foo" "bar" FROM "abc"', str(q))

    def test_select__column__single__table_alias__str(self):
        q = Query.from_(self.table_abc.as_("fizzbuzz")).select(self.table_abc.foo.as_("bar"))

        self.assertEqual('SELECT "foo" "bar" FROM "abc" "fizzbuzz"', str(q))

    def test_select__column__single__field(self):
        t = Table("abc")
        q = Query.from_(t).select(t.foo)

        self.assertEqual('SELECT "foo" FROM "abc"', str(q))

    def test_select__columns__multi__str(self):
        q1 = Query.from_("abc").select("foo", "bar")
        q2 = Query.from_("abc").select("foo").select("bar")

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q1))
        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q2))

    def test_select__columns__multi__field(self):
        q1 = Query.from_(self.table_abc).select(self.table_abc.foo, self.table_abc.bar)
        q2 = Query.from_(self.table_abc).select(self.table_abc.foo).select(self.table_abc.bar)

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q1))
        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q2))

    def test_select__multiple_tables(self):
        q = Query.from_(self.table_abc).select(self.table_abc.foo).from_(self.table_efg).select(self.table_efg.bar)

        self.assertEqual('SELECT "abc"."foo","efg"."bar" FROM "abc","efg"', str(q))

    def test_select__subquery(self):
        subquery = Query.from_(self.table_abc).select("*")
        q = Query.from_(subquery).select(subquery.foo, subquery.bar)

        self.assertEqual('SELECT "sq0"."foo","sq0"."bar" FROM (SELECT * FROM "abc") "sq0"', str(q))

    def test_select__multiple_subqueries(self):
        subquery0 = Query.from_(self.table_abc).select("foo")
        subquery1 = Query.from_(self.table_efg).select("bar")
        q = Query.from_(subquery0).from_(subquery1).select(subquery0.foo, subquery1.bar)

        self.assertEqual(
            'SELECT "sq0"."foo","sq1"."bar" FROM (SELECT "foo" FROM "abc") "sq0",(SELECT "bar" FROM "efg") "sq1"',
            str(q),
        )

    def test_select__nested_subquery(self):
        subquery0 = Query.from_(self.table_abc).select("*")
        subquery1 = Query.from_(subquery0).select(subquery0.foo, subquery0.bar)
        subquery2 = Query.from_(subquery1).select(subquery1.foo)

        q = Query.from_(subquery2).select(subquery2.foo)

        self.assertEqual(
            'SELECT "sq2"."foo" '
            'FROM (SELECT "sq1"."foo" '
            'FROM (SELECT "sq0"."foo","sq0"."bar" '
            'FROM (SELECT * FROM "abc") "sq0") "sq1") "sq2"',
            str(q),
        )

    def test_select__no_table(self):
        q = Query.select(1, 2, 3)

        self.assertEqual("SELECT 1,2,3", str(q))

    def test_select_then_add_table(self):
        q = Query.select(1).select(2, 3).from_("abc").select("foo")

        self.assertEqual('SELECT 1,2,3,"foo" FROM "abc"', str(q))

    def test_select_with_limit(self):
        q1 = Query.from_("abc").select("foo")[:10]

        self.assertEqual('SELECT "foo" FROM "abc" LIMIT 10', str(q1))

    def test_select_with_limit_zero(self):
        with self.subTest('using python slice'):
            q1 = Query.from_("abc").select("foo")[:0]
            self.assertEqual('SELECT "foo" FROM "abc" LIMIT 0', str(q1))

        with self.subTest('using limit method'):
            q2 = Query.from_("abc").select("foo").limit(0)
            self.assertEqual('SELECT "foo" FROM "abc" LIMIT 0', str(q2))

    def test_select_with_limit__func(self):
        q1 = Query.from_("abc").select("foo").limit(10)

        self.assertEqual('SELECT "foo" FROM "abc" LIMIT 10', str(q1))

    def test_select_with_offset(self):
        q1 = Query.from_("abc").select("foo")[10:]

        self.assertEqual('SELECT "foo" FROM "abc" OFFSET 10', str(q1))

    def test_select_with_offset__func(self):
        q1 = Query.from_("abc").select("foo").offset(10)

        self.assertEqual('SELECT "foo" FROM "abc" OFFSET 10', str(q1))

    def test_select_with_limit_and_offset(self):
        q1 = Query.from_("abc").select("foo")[10:10]

        self.assertEqual('SELECT "foo" FROM "abc" LIMIT 10 OFFSET 10', str(q1))

    def test_select_with_force_index(self):
        q = Query.from_("abc").select("foo").force_index("egg")

        self.assertEqual('SELECT "foo" FROM "abc" FORCE INDEX ("egg")', str(q))

    def test_select_with_force_index_with_index_object(self):
        index = Index("egg")
        q = Query.from_("abc").select("foo").force_index(index)

        self.assertEqual('SELECT "foo" FROM "abc" FORCE INDEX ("egg")', str(q))

    def test_select_with_force_index_multiple_indexes(self):
        q = Query.from_("abc").select("foo").force_index("egg", "bacon")

        self.assertEqual('SELECT "foo" FROM "abc" FORCE INDEX ("egg","bacon")', str(q))

    def test_select_with_force_index_multiple_calls(self):
        q = (
            Query.from_("abc")
            .select("foo")
            .force_index(
                "egg",
            )
            .force_index("spam")
        )

        self.assertEqual('SELECT "foo" FROM "abc" FORCE INDEX ("egg","spam")', str(q))

    def test_select_with_use_index(self):
        q = Query.from_("abc").select("foo").use_index("egg")

        self.assertEqual('SELECT "foo" FROM "abc" USE INDEX ("egg")', str(q))

    def test_select_with_use_index_with_index_object(self):
        index = Index("egg")
        q = Query.from_("abc").select("foo").use_index(index)

        self.assertEqual('SELECT "foo" FROM "abc" USE INDEX ("egg")', str(q))

    def test_select_with_use_index_multiple_indexes(self):
        q = Query.from_("abc").select("foo").use_index("egg", "bacon")

        self.assertEqual('SELECT "foo" FROM "abc" USE INDEX ("egg","bacon")', str(q))

    def test_select_with_use_index_multiple_calls(self):
        q = (
            Query.from_("abc")
            .select("foo")
            .use_index(
                "egg",
            )
            .use_index("spam")
        )

        self.assertEqual('SELECT "foo" FROM "abc" USE INDEX ("egg","spam")', str(q))

    def test_mysql_query_uses_backtick_quote_chars(self):
        q = MySQLQuery.from_("abc").select("foo", "bar")

        self.assertEqual("SELECT `foo`,`bar` FROM `abc`", str(q))

    def test_vertica_query_uses_double_quote_chars(self):
        q = VerticaQuery.from_("abc").select("foo", "bar")

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q))

    def test_mssql_query_uses_double_quote_chars(self):
        q = MSSQLQuery.from_("abc").select("foo", "bar")

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q))

    def test_oracle_query_uses_no_quote_chars(self):
        q = OracleQuery.from_("abc").select("foo", "bar")

        self.assertEqual('SELECT foo,bar FROM abc', str(q))

    def test_postgresql_query_uses_double_quote_chars(self):
        q = PostgreSQLQuery.from_("abc").select("foo", "bar")

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q))

    def test_redshift_query_uses_double_quote_chars(self):
        q = RedshiftQuery.from_("abc").select("foo", "bar")

        self.assertEqual('SELECT "foo","bar" FROM "abc"', str(q))

    def test_table_select_alias(self):
        q = self.table_abc.select(1)

        self.assertEqual('SELECT 1 FROM "abc"', str(q))
        self.assertEqual(q, Query.from_("abc").select(1))

    def test_table_select_alias_with_offset_and_limit(self):
        self.assertEqual(self.table_abc.select("foo")[10:10], Query.from_("abc").select("foo")[10:10])
        self.assertEqual(
            self.table_abc.select(self.table_abc.foo)[10:10],
            Query.from_("abc").select("foo")[10:10],
        )

    def test_temporal_select(self):
        t = Table("abc")

        with self.subTest("with system time as of"):
            q = Query.from_(t.for_(SYSTEM_TIME.as_of('2020-01-01'))).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR SYSTEM_TIME AS OF \'2020-01-01\'', str(q))

        with self.subTest("with system time between"):
            q = Query.from_(t.for_(SYSTEM_TIME.between('2020-01-01', '2020-02-01'))).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR SYSTEM_TIME BETWEEN \'2020-01-01\' AND \'2020-02-01\'', str(q))

        with self.subTest("with system time from to"):
            q = Query.from_(t.for_(SYSTEM_TIME.from_to('2020-01-01', '2020-02-01'))).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR SYSTEM_TIME FROM \'2020-01-01\' TO \'2020-02-01\'', str(q))

        with self.subTest("with ALL"):
            q = Query.from_(t.for_(SYSTEM_TIME.all_())).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR SYSTEM_TIME ALL', str(q))

        with self.subTest("with period between"):
            q = Query.from_(t.for_(t.valid_period.between('2020-01-01', '2020-02-01'))).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR "valid_period" BETWEEN \'2020-01-01\' AND \'2020-02-01\'', str(q))

        with self.subTest("with period from to"):
            q = Query.from_(t.for_(t.valid_period.from_to('2020-01-01', '2020-02-01'))).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR "valid_period" FROM \'2020-01-01\' TO \'2020-02-01\'', str(q))

        with self.subTest("with ALL"):
            q = Query.from_(t.for_(t.valid_period.all_())).select("*")

            self.assertEqual('SELECT * FROM "abc" FOR "valid_period" ALL', str(q))


class MyEnum(Enum):
    STR = "foo"
    INT = 0
    BOOL = True
    DATE = date(2020, 2, 2)
    NONE = None


class WhereTests(unittest.TestCase):
    t = Table("abc")
    t2 = Table("cba")

    def test_where_enum(self):
        q1 = Query.from_(self.t).select("*").where(self.t.foo == MyEnum.STR)
        q2 = Query.from_(self.t).select("*").where(self.t.foo == MyEnum.INT)
        q3 = Query.from_(self.t).select("*").where(self.t.foo == MyEnum.BOOL)
        q4 = Query.from_(self.t).select("*").where(self.t.foo == MyEnum.DATE)
        q5 = Query.from_(self.t).select("*").where(self.t.foo == MyEnum.NONE)

        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=\'foo\'', str(q1))
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=0', str(q2))
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=true', str(q3))
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=\'2020-02-02\'', str(q4))
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=null', str(q5))

    def test_where_field_equals(self):
        q1 = Query.from_(self.t).select("*").where(self.t.foo == self.t.bar)
        q2 = Query.from_(self.t).select("*").where(self.t.foo.eq(self.t.bar))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo"="bar"', str(q1))
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"="bar"', str(q2))
        q = self.t.select("*").where(self.t.foo == self.t.bar)
        self.assertEqual(q, q1)

    def test_where_field_equals_for_update(self):
        q = Query.from_(self.t).select("*").where(self.t.foo == self.t.bar).for_update()
        self.assertEqual('SELECT * FROM "abc" WHERE "foo"="bar" FOR UPDATE', str(q))

    def test_where_field_equals_for_update_nowait(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            q = query_cls.from_(self.t).select("*").where(self.t.foo == self.t.bar).for_update(nowait=True)
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                'FOR UPDATE NOWAIT',
                str(q),
            )

    def test_where_field_equals_for_update_skip_locked(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            q = query_cls.from_(self.t).select("*").where(self.t.foo == self.t.bar).for_update(skip_locked=True)
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                'FOR UPDATE SKIP LOCKED',
                str(q),
            )

    def test_where_field_equals_for_update_of(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            q = query_cls.from_(self.t).select("*").where(self.t.foo == self.t.bar).for_update(of=("abc",))
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                f'FOR UPDATE OF {quote_char}abc{quote_char}',
                str(q),
            )

    def test_where_field_equals_for_update_of_multiple_tables(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            q = (
                query_cls.from_(self.t)
                .join(self.t2)
                .on(self.t.id == self.t2.abc_id)
                .select("*")
                .where(self.t.foo == self.t.bar)
                .for_update(of=("abc", "cba"))
            )
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            self.assertIn(
                str(q),
                [
                    'SELECT * '
                    f'FROM {quote_char}abc{quote_char} '
                    f'JOIN {quote_char}cba{quote_char} '
                    f'ON {quote_char}abc{quote_char}.{quote_char}id{quote_char}='
                    f'{quote_char}cba{quote_char}.{quote_char}abc_id{quote_char} '
                    f'WHERE {quote_char}abc{quote_char}.{quote_char}foo{quote_char}='
                    f'{quote_char}abc{quote_char}.{quote_char}bar{quote_char} '
                    f'FOR UPDATE OF {quote_char}cba{quote_char}, {quote_char}abc{quote_char}',
                    'SELECT * '
                    f'FROM {quote_char}abc{quote_char} '
                    f'JOIN {quote_char}cba{quote_char} '
                    f'ON {quote_char}abc{quote_char}.{quote_char}id{quote_char}='
                    f'{quote_char}cba{quote_char}.{quote_char}abc_id{quote_char} '
                    f'WHERE {quote_char}abc{quote_char}.{quote_char}foo{quote_char}='
                    f'{quote_char}abc{quote_char}.{quote_char}bar{quote_char} '
                    f'FOR UPDATE OF {quote_char}abc{quote_char}, {quote_char}cba{quote_char}',
                ],
            )

    def test_where_field_equals_for_update_of_nowait(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            q = query_cls.from_(self.t).select("*").where(self.t.foo == self.t.bar).for_update(of=("abc",), nowait=True)
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                f'FOR UPDATE OF {quote_char}abc{quote_char} NOWAIT',
                str(q),
            )

    def test_where_field_equals_for_update_of_skip_locked(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            q = (
                query_cls.from_(self.t)
                .select("*")
                .where(self.t.foo == self.t.bar)
                .for_update(of=("abc",), skip_locked=True)
            )
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                f'FOR UPDATE OF {quote_char}abc{quote_char} SKIP LOCKED',
                str(q),
            )

    def test_where_field_equals_for_update_skip_locked_and_of(self):
        for query_cls in [
            MySQLQuery,
            PostgreSQLQuery,
        ]:
            q = (
                query_cls.from_(self.t)
                .select("*")
                .where(self.t.foo == self.t.bar)
                .for_update(nowait=False, skip_locked=True, of=("abc",))
            )
            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'
            self.assertEqual(
                'SELECT * '
                f'FROM {quote_char}abc{quote_char} '
                f'WHERE {quote_char}foo{quote_char}={quote_char}bar{quote_char} '
                f'FOR UPDATE OF {quote_char}abc{quote_char} SKIP LOCKED',
                str(q),
            )

    def test_where_field_equals_where(self):
        q = Query.from_(self.t).select("*").where(self.t.foo == 1).where(self.t.bar == self.t.baz)

        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=1 AND "bar"="baz"', str(q))

    def test_where_field_equals_where_not(self):
        q = Query.from_(self.t).select("*").where((self.t.foo == 1).negate()).where(self.t.bar == self.t.baz)

        self.assertEqual('SELECT * FROM "abc" WHERE NOT "foo"=1 AND "bar"="baz"', str(q))

    def test_where_field_equals_where_two_not(self):
        q = Query.from_(self.t).select("*").where((self.t.foo == 1).negate()).where((self.t.bar == self.t.baz).negate())

        self.assertEqual('SELECT * FROM "abc" WHERE NOT "foo"=1 AND NOT "bar"="baz"', str(q))

    def test_where_single_quote(self):
        q1 = Query.from_(self.t).select("*").where(self.t.foo == "bar'foo")

        self.assertEqual("SELECT * FROM \"abc\" WHERE \"foo\"='bar''foo'", str(q1))

    def test_where_field_equals_and(self):
        q = Query.from_(self.t).select("*").where((self.t.foo == 1) & (self.t.bar == self.t.baz))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=1 AND "bar"="baz"', str(q))

    def test_where_field_equals_or(self):
        q = Query.from_(self.t).select("*").where((self.t.foo == 1) | (self.t.bar == self.t.baz))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo"=1 OR "bar"="baz"', str(q))

    def test_where_nested_conditions(self):
        q = Query.from_(self.t).select("*").where((self.t.foo == 1) | (self.t.bar == self.t.baz)).where(self.t.baz == 0)

        self.assertEqual('SELECT * FROM "abc" WHERE ("foo"=1 OR "bar"="baz") AND "baz"=0', str(q))

    def test_where_field_starts_with(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.like("ab%"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" LIKE \'ab%\'', str(q))

    def test_where_field_contains(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.like("%fg%"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" LIKE \'%fg%\'', str(q))

    def test_where_field_ends_with(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.like("%yz"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" LIKE \'%yz\'', str(q))

    def test_where_field_is_n_chars_long(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.like("___"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" LIKE \'___\'', str(q))

    def test_where_field_does_not_start_with(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.not_like("ab%"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" NOT LIKE \'ab%\'', str(q))

    def test_where_field_does_not_contain(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.not_like("%fg%"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" NOT LIKE \'%fg%\'', str(q))

    def test_where_field_does_not_end_with(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.not_like("%yz"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" NOT LIKE \'%yz\'', str(q))

    def test_where_field_is_not_n_chars_long(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.not_like("___"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" NOT LIKE \'___\'', str(q))

    def test_where_field_matches_regex(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.regex(r"^b"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" REGEX \'^b\'', str(q))

    def test_where_field_matches_regexp(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.regexp(r"^b"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" REGEXP \'^b\'', str(q))

    def test_where_field_matches_rlike(self):
        q = Query.from_(self.t).select(self.t.star).where(self.t.foo.rlike(r"^b"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" RLIKE \'^b\'', str(q))

    def test_ignore_empty_criterion_where(self):
        q1 = Query.from_(self.t).select("*").where(EmptyCriterion())

        self.assertEqual('SELECT * FROM "abc"', str(q1))

    def test_ignore_empty_criterion_having(self):
        q1 = Query.from_(self.t).select("*").having(EmptyCriterion())

        self.assertEqual('SELECT * FROM "abc"', str(q1))

    def test_select_with_force_index_and_where(self):
        q = Query.from_("abc").select("foo").where(self.t.foo == self.t.bar).force_index("egg")

        self.assertEqual('SELECT "foo" FROM "abc" FORCE INDEX ("egg") WHERE "foo"="bar"', str(q))

    def test_where_with_multiple_wheres_using_and_case(self):
        case_stmt = Case().when(self.t.foo == 'bar', 1).else_(0)
        query = Query.from_(self.t).select(case_stmt).where(case_stmt & self.t.blah.isin(['test']))

        self.assertEqual(
            'SELECT CASE WHEN "foo"=\'bar\' THEN 1 ELSE 0 END FROM "abc" WHERE CASE WHEN "foo"=\'bar\' THEN 1 ELSE 0 '
            'END AND "blah" IN (\'test\')',
            str(query),
        )

    def test_where_with_multiple_wheres_using_or_case(self):
        case_stmt = Case().when(self.t.foo == 'bar', 1).else_(0)
        query = Query.from_(self.t).select(case_stmt).where(case_stmt | self.t.blah.isin(['test']))

        self.assertEqual(
            'SELECT CASE WHEN "foo"=\'bar\' THEN 1 ELSE 0 END FROM "abc" WHERE CASE WHEN "foo"=\'bar\' THEN 1 ELSE 0 '
            'END OR "blah" IN (\'test\')',
            str(query),
        )


class PreWhereTests(WhereTests):
    t = Table("abc")

    def test_prewhere_field_equals(self):
        q1 = Query.from_(self.t).select("*").prewhere(self.t.foo == self.t.bar)
        q2 = Query.from_(self.t).select("*").prewhere(self.t.foo.eq(self.t.bar))

        self.assertEqual('SELECT * FROM "abc" PREWHERE "foo"="bar"', str(q1))
        self.assertEqual('SELECT * FROM "abc" PREWHERE "foo"="bar"', str(q2))

    def test_where_and_prewhere(self):
        q = Query.from_(self.t).select("*").prewhere(self.t.foo == self.t.bar).where(self.t.foo == self.t.bar)

        self.assertEqual('SELECT * FROM "abc" PREWHERE "foo"="bar" WHERE "foo"="bar"', str(q))


class GroupByTests(unittest.TestCase):
    t = Table("abc")
    maxDiff = None

    def test_groupby__single(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__multi(self):
        q = Query.from_(self.t).groupby(self.t.foo, self.t.bar).select(self.t.foo, self.t.bar)

        self.assertEqual('SELECT "foo","bar" FROM "abc" GROUP BY "foo","bar"', str(q))

    def test_groupby__count_star(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo, fn.Count("*"))

        self.assertEqual('SELECT "foo",COUNT(*) FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__count_field(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo, fn.Count(self.t.bar))

        self.assertEqual('SELECT "foo",COUNT("bar") FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__count_distinct(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo, fn.Count("*").distinct())

        self.assertEqual('SELECT "foo",COUNT(DISTINCT *) FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__sum_distinct(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo, fn.Sum(self.t.bar).distinct())

        self.assertEqual('SELECT "foo",SUM(DISTINCT "bar") FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__sum_filter(self):
        q = (
            Query.from_(self.t)
            .groupby(self.t.foo)
            .select(self.t.foo, fn.Sum(self.t.bar).filter(self.t.id.eq(1) & self.t.cid.gt(2)))
        )

        self.assertEqual('SELECT "foo",SUM("bar") FILTER(WHERE "id"=1 AND "cid">2) FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__str(self):
        q = Query.from_("abc").groupby("foo").select("foo", fn.Count("*").distinct())

        self.assertEqual('SELECT "foo",COUNT(DISTINCT *) FROM "abc" GROUP BY "foo"', str(q))

    def test_groupby__int(self):
        q = Query.from_("abc").groupby(1).select("foo", fn.Count("*").distinct())

        self.assertEqual('SELECT "foo",COUNT(DISTINCT *) FROM "abc" GROUP BY 1', str(q))

    def test_groupby__alias(self):
        bar = self.t.bar.as_("bar01")
        q = Query.from_(self.t).select(fn.Sum(self.t.foo), bar).groupby(bar)

        self.assertEqual('SELECT SUM("foo"),"bar" "bar01" FROM "abc" GROUP BY "bar01"', str(q))

    def test_groupby__no_alias(self):
        bar = self.t.bar.as_("bar01")
        q = Query.from_(self.t).select(fn.Sum(self.t.foo), bar).groupby(bar)

        self.assertEqual(
            'SELECT SUM("foo"),"bar" "bar01" FROM "abc" GROUP BY "bar"',
            q.get_sql(groupby_alias=False),
        )

    def test_groupby__no_alias_mssql(self):
        bar = self.t.bar.as_("bar01")
        q = MSSQLQuery.from_(self.t).select(fn.Sum(self.t.foo), bar).groupby(bar)

        self.assertEqual('SELECT SUM("foo"),"bar" "bar01" FROM "abc" GROUP BY "bar"', str(q))

    def test_groupby__no_alias_oracle(self):
        bar = self.t.bar.as_("bar01")
        q = OracleQuery.from_(self.t).select(fn.Sum(self.t.foo), bar).groupby(bar)

        self.assertEqual('SELECT SUM(foo),bar bar01 FROM abc GROUP BY bar', str(q))

    def test_groupby__alias_platforms(self):
        bar = self.t.bar.as_("bar01")

        for query_cls in [
            MySQLQuery,
            VerticaQuery,
            PostgreSQLQuery,
            RedshiftQuery,
            ClickHouseQuery,
            SQLLiteQuery,
        ]:
            q = query_cls.from_(self.t).select(fn.Sum(self.t.foo), bar).groupby(bar)

            quote_char = query_cls._builder().QUOTE_CHAR if isinstance(query_cls._builder().QUOTE_CHAR, str) else '"'

            self.assertEqual(
                "SELECT "
                "SUM({quote_char}foo{quote_char}),"
                "{quote_char}bar{quote_char}{as_keyword}{quote_char}bar01{quote_char} "
                "FROM {quote_char}abc{quote_char} "
                "GROUP BY {quote_char}bar01{quote_char}".format(
                    as_keyword=' AS ' if query_cls is ClickHouseQuery else ' ',
                    quote_char=quote_char,
                ),
                str(q),
            )

    def test_groupby__alias_with_join(self):
        table1 = Table("table1", alias="t1")
        bar = table1.bar.as_("bar01")
        q = Query.from_(self.t).join(table1).on(self.t.id == table1.t_ref).select(fn.Sum(self.t.foo), bar).groupby(bar)

        self.assertEqual(
            'SELECT SUM("abc"."foo"),"t1"."bar" "bar01" FROM "abc" '
            'JOIN "table1" "t1" ON "abc"."id"="t1"."t_ref" '
            'GROUP BY "bar01"',
            str(q),
        )

    def test_groupby_with_case_uses_the_alias(self):
        q = (
            Query.from_(self.t)
            .select(
                fn.Sum(self.t.foo).as_("bar"),
                Case().when(self.t.fname == "Tom", "It was Tom").else_("It was someone else.").as_("who_was_it"),
            )
            .groupby(Case().when(self.t.fname == "Tom", "It was Tom").else_("It was someone else.").as_("who_was_it"))
        )

        self.assertEqual(
            'SELECT SUM("foo") "bar",'
            "CASE WHEN \"fname\"='Tom' THEN 'It was Tom' "
            "ELSE 'It was someone else.' END \"who_was_it\" "
            'FROM "abc" '
            'GROUP BY "who_was_it"',
            str(q),
        )

    def test_mysql_query_uses_backtick_quote_chars(self):
        q = MySQLQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual("SELECT `foo` FROM `abc` GROUP BY `foo`", str(q))

    def test_vertica_query_uses_double_quote_chars(self):
        q = VerticaQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_mssql_query_uses_double_quote_chars(self):
        q = MSSQLQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_oracle_query_uses_no_quote_chars(self):
        q = OracleQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT foo FROM abc GROUP BY foo', str(q))

    def test_postgres_query_uses_double_quote_chars(self):
        q = PostgreSQLQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_redshift_query_uses_double_quote_chars(self):
        q = RedshiftQuery.from_(self.t).groupby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_group_by__single_with_totals(self):
        q = Query.from_(self.t).groupby(self.t.foo).select(self.t.foo).with_totals()

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo" WITH TOTALS', str(q))

    def test_groupby__multi_with_totals(self):
        q = Query.from_(self.t).groupby(self.t.foo, self.t.bar).select(self.t.foo, self.t.bar).with_totals()

        self.assertEqual('SELECT "foo","bar" FROM "abc" GROUP BY "foo","bar" WITH TOTALS', str(q))


class HavingTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_having_greater_than(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, fn.Sum(self.table_abc.bar))
            .groupby(self.table_abc.foo)
            .having(fn.Sum(self.table_abc.bar) > 1)
        )

        self.assertEqual(
            'SELECT "foo",SUM("bar") FROM "abc" GROUP BY "foo" HAVING SUM("bar")>1',
            str(q),
        )

    def test_having_and(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, fn.Sum(self.table_abc.bar))
            .groupby(self.table_abc.foo)
            .having((fn.Sum(self.table_abc.bar) > 1) & (fn.Sum(self.table_abc.bar) < 100))
        )

        self.assertEqual(
            'SELECT "foo",SUM("bar") FROM "abc" GROUP BY "foo" HAVING SUM("bar")>1 AND SUM("bar")<100',
            str(q),
        )

    def test_having_join_and_equality(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.foo)
            .select(self.table_abc.foo, fn.Sum(self.table_efg.bar), self.table_abc.buz)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
            .having(fn.Sum(self.table_efg.bar) > 100)
        )

        self.assertEqual(
            'SELECT "abc"."foo",SUM("efg"."bar"),"abc"."buz" FROM "abc" '
            'JOIN "efg" ON "abc"."foo"="efg"."foo" '
            'GROUP BY "abc"."foo" '
            'HAVING "abc"."buz"=\'fiz\' AND SUM("efg"."bar")>100',
            str(q),
        )

    def test_mysql_query_uses_backtick_quote_chars(self):
        q = (
            MySQLQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual("SELECT `foo` FROM `abc` GROUP BY `foo` HAVING `buz`='fiz'", str(q))

    def test_vertica_query_uses_double_quote_chars(self):
        q = (
            VerticaQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo" HAVING "buz"=\'fiz\'', str(q))

    def test_mssql_query_uses_double_quote_chars(self):
        q = (
            MSSQLQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo" HAVING "buz"=\'fiz\'', str(q))

    def test_oracle_query_uses_no_quote_chars(self):
        q = (
            OracleQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual('SELECT foo FROM abc GROUP BY foo HAVING buz=\'fiz\'', str(q))

    def test_postgres_query_uses_double_quote_chars(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo" HAVING "buz"=\'fiz\'', str(q))

    def test_redshift_query_uses_double_quote_chars(self):
        q = (
            RedshiftQuery.from_(self.table_abc)
            .select(self.table_abc.foo)
            .groupby(self.table_abc.foo)
            .having(self.table_abc.buz == "fiz")
        )
        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo" HAVING "buz"=\'fiz\'', str(q))


class OrderByTests(unittest.TestCase):
    t = Table("abc")

    def test_orderby_single_field(self):
        q = Query.from_(self.t).orderby(self.t.foo).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" ORDER BY "foo"', str(q))

    def test_orderby_multi_fields(self):
        q = Query.from_(self.t).orderby(self.t.foo, self.t.bar).select(self.t.foo, self.t.bar)

        self.assertEqual('SELECT "foo","bar" FROM "abc" ORDER BY "foo","bar"', str(q))

    def test_orderby_single_str(self):
        q = Query.from_("abc").orderby("foo").select("foo")

        self.assertEqual('SELECT "foo" FROM "abc" ORDER BY "foo"', str(q))

    def test_orderby_asc(self):
        q = Query.from_(self.t).orderby(self.t.foo, order=Order.asc).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" ORDER BY "foo" ASC', str(q))

    def test_orderby_desc(self):
        q = Query.from_(self.t).orderby(self.t.foo, order=Order.desc).select(self.t.foo)

        self.assertEqual('SELECT "foo" FROM "abc" ORDER BY "foo" DESC', str(q))

    def test_orderby_no_alias(self):
        bar = self.t.bar.as_("bar01")
        q = Query.from_(self.t).select(fn.Sum(self.t.foo), bar).orderby(bar)

        self.assertEqual(
            'SELECT SUM("foo"),"bar" "bar01" FROM "abc" ORDER BY "bar"',
            q.get_sql(orderby_alias=False),
        )

    def test_orderby_alias(self):
        bar = self.t.bar.as_("bar01")
        q = Query.from_(self.t).select(fn.Sum(self.t.foo), bar).orderby(bar)

        self.assertEqual('SELECT SUM("foo"),"bar" "bar01" FROM "abc" ORDER BY "bar01"', q.get_sql())


class AliasTests(unittest.TestCase):
    t = Table("abc")

    def test_table_field(self):
        q = Query.from_(self.t).select(self.t.foo.as_("bar"))

        self.assertEqual('SELECT "foo" "bar" FROM "abc"', str(q))

    def test_table_field__multi(self):
        q = Query.from_(self.t).select(self.t.foo.as_("bar"), self.t.fiz.as_("buz"))

        self.assertEqual('SELECT "foo" "bar","fiz" "buz" FROM "abc"', str(q))

    def test_arithmetic_function(self):
        q = Query.from_(self.t).select((self.t.foo + self.t.bar).as_("biz"))

        self.assertEqual('SELECT "foo"+"bar" "biz" FROM "abc"', str(q))

    def test_functions_using_as(self):
        q = Query.from_(self.t).select(fn.Count("*").as_("foo"))

        self.assertEqual('SELECT COUNT(*) "foo" FROM "abc"', str(q))

    def test_functions_using_constructor_param(self):
        q = Query.from_(self.t).select(fn.Count("*", alias="foo"))

        self.assertEqual('SELECT COUNT(*) "foo" FROM "abc"', str(q))

    def test_function_using_as_nested(self):
        """
        We don't show aliases of fields that are arguments of a function.
        """
        q = Query.from_(self.t).select(fn.Sqrt(fn.Count("*").as_("foo")).as_("bar"))

        self.assertEqual('SELECT SQRT(COUNT(*)) "bar" FROM "abc"', str(q))

    def test_functions_using_constructor_param_nested(self):
        """
        We don't show aliases of fields that are arguments of a function.
        """
        q = Query.from_(self.t).select(fn.Sqrt(fn.Count("*", alias="foo"), alias="bar"))

        self.assertEqual('SELECT SQRT(COUNT(*)) "bar" FROM "abc"', str(q))

    def test_ignored_in_where(self):
        q = Query.from_(self.t).select(self.t.foo).where(self.t.foo.as_("bar") == 1)

        self.assertEqual('SELECT "foo" FROM "abc" WHERE "foo"=1', str(q))

    def test_ignored_in_groupby(self):
        q = Query.from_(self.t).select(self.t.foo).groupby(self.t.foo.as_("bar"))

        self.assertEqual('SELECT "foo" FROM "abc" GROUP BY "foo"', str(q))

    def test_ignored_in_orderby(self):
        q = Query.from_(self.t).select(self.t.foo).orderby(self.t.foo.as_("bar"))

        self.assertEqual('SELECT "foo" FROM "abc" ORDER BY "foo"', str(q))

    def test_ignored_in_criterion(self):
        c = self.t.foo.as_("bar") == 1

        self.assertEqual('"foo"=1', str(c))

    def test_ignored_in_criterion_comparison(self):
        c = self.t.foo.as_("bar") == self.t.fiz.as_("buz")

        self.assertEqual('"foo"="fiz"', str(c))

    def test_ignored_in_field_inside_case(self):
        q = Query.from_(self.t).select(Case().when(self.t.foo == 1, "a").else_(self.t.bar.as_('"buz"')))

        self.assertEqual('SELECT CASE WHEN "foo"=1 THEN \'a\' ELSE "bar" END FROM "abc"', str(q))

    def test_case_using_as(self):
        q = Query.from_(self.t).select(Case().when(self.t.foo == 1, "a").else_("b").as_("bar"))

        self.assertEqual(
            'SELECT CASE WHEN "foo"=1 THEN \'a\' ELSE \'b\' END "bar" FROM "abc"',
            str(q),
        )

    def test_case_using_constructor_param(self):
        q = Query.from_(self.t).select(Case(alias="bar").when(self.t.foo == 1, "a").else_("b"))

        self.assertEqual(
            'SELECT CASE WHEN "foo"=1 THEN \'a\' ELSE \'b\' END "bar" FROM "abc"',
            str(q),
        )

    def test_select__multiple_tables(self):
        table_abc, table_efg = Table("abc", alias="q0"), Table("efg", alias="q1")

        q = Query.from_(table_abc).select(table_abc.foo).from_(table_efg).select(table_efg.bar)

        self.assertEqual('SELECT "q0"."foo","q1"."bar" FROM "abc" "q0","efg" "q1"', str(q))

    def test_use_aliases_in_groupby_and_orderby(self):
        table_abc = Table("abc", alias="q0")

        my_foo = table_abc.foo.as_("my_foo")
        q = Query.from_(table_abc).select(my_foo, table_abc.bar).groupby(my_foo).orderby(my_foo)

        self.assertEqual(
            'SELECT "q0"."foo" "my_foo","q0"."bar" FROM "abc" "q0" GROUP BY "my_foo" ORDER BY "my_foo"',
            str(q),
        )

    def test_table_with_schema_and_alias(self):
        table = Table("abc", schema="schema", alias="alias")
        self.assertEqual('"schema"."abc" "alias"', str(table))

    def test_null_value_with_alias(self):
        q = Query.select(NullValue().as_("abcdef"))

        self.assertEqual('SELECT NULL "abcdef"', str(q))


class SubqueryTests(unittest.TestCase):
    maxDiff = None

    table_abc, table_efg, table_hij = Tables("abc", "efg", "hij")

    def test_where__in(self):
        q = (
            Query.from_(self.table_abc)
            .select("*")
            .where(
                self.table_abc.foo.isin(
                    Query.from_(self.table_efg).select(self.table_efg.foo).where(self.table_efg.bar == 0),
                ),
            )
        )

        self.assertEqual(
            'SELECT * FROM "abc" WHERE "foo" IN (SELECT "foo" FROM "efg" WHERE "bar"=0)',
            str(q),
        )

    def test_where__in_nested(self):
        q = Query.from_(self.table_abc).select("*").where(self.table_abc.foo).isin(self.table_efg.select("*"))

        self.assertEqual('SELECT * FROM "abc" WHERE "foo" IN (SELECT * FROM "efg")', str(q))

    def test_join(self):
        subquery = Query.from_("efg").select("fiz", "buz").where(F("buz") == 0)

        q = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(self.table_abc.bar == subquery.buz)
            .select(self.table_abc.foo, subquery.fiz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","sq0"."fiz" FROM "abc" '
            'JOIN (SELECT "fiz","buz" FROM "efg" WHERE "buz"=0) "sq0" '
            'ON "abc"."bar"="sq0"."buz"',
            str(q),
        )

    def test_select_subquery(self):
        subq = Query.from_(self.table_efg).select("fizzbuzz").where(self.table_efg.id == 1)

        q = Query.from_(self.table_abc).select("foo", "bar").select(subq)

        self.assertEqual(
            'SELECT "foo","bar",(SELECT "fizzbuzz" FROM "efg" WHERE "id"=1) FROM "abc"',
            str(q),
        )

    def test_select_subquery_with_alias(self):
        subq = Query.from_(self.table_efg).select("fizzbuzz").where(self.table_efg.id == 1)

        q = Query.from_(self.table_abc).select("foo", "bar").select(subq.as_("sq"))

        self.assertEqual(
            'SELECT "foo","bar",(SELECT "fizzbuzz" FROM "efg" WHERE "id"=1) "sq" FROM "abc"',
            str(q),
        )

    def test_where__equality(self):
        subquery = Query.from_("efg").select("fiz").where(F("buz") == 0)
        query = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(self.table_abc.bar == subquery)
        )

        self.assertEqual(
            'SELECT "foo","bar" FROM "abc" WHERE "bar"=(SELECT "fiz" FROM "efg" WHERE "buz"=0)',
            str(query),
        )

    def test_select_from_nested_query(self):
        subquery = Query.from_(self.table_abc).select(
            self.table_abc.foo,
            self.table_abc.bar,
            (self.table_abc.fizz + self.table_abc.buzz).as_("fizzbuzz"),
        )

        query = Query.from_(subquery).select(subquery.foo, subquery.bar, subquery.fizzbuzz)

        self.assertEqual(
            'SELECT "sq0"."foo","sq0"."bar","sq0"."fizzbuzz" '
            "FROM ("
            'SELECT "foo","bar","fizz"+"buzz" "fizzbuzz" '
            'FROM "abc"'
            ') "sq0"',
            str(query),
        )

    def test_select_from_nested_query_with_join(self):
        subquery1 = (
            Query.from_(self.table_abc)
            .select(
                self.table_abc.foo,
                fn.Sum(self.table_abc.fizz + self.table_abc.buzz).as_("fizzbuzz"),
            )
            .groupby(self.table_abc.foo)
        )

        subquery2 = Query.from_(self.table_efg).select(
            self.table_efg.foo.as_("foo_two"),
            self.table_efg.bar,
        )

        query = (
            Query.from_(subquery1)
            .select(subquery1.foo, subquery1.fizzbuzz)
            .join(subquery2)
            .on(subquery1.foo == subquery2.foo_two)
            .select(subquery2.foo_two, subquery2.bar)
        )

        self.assertEqual(
            "SELECT "
            '"sq0"."foo","sq0"."fizzbuzz",'
            '"sq1"."foo_two","sq1"."bar" '
            "FROM ("
            "SELECT "
            '"foo",SUM("fizz"+"buzz") "fizzbuzz" '
            'FROM "abc" '
            'GROUP BY "foo"'
            ') "sq0" JOIN ('
            "SELECT "
            '"foo" "foo_two","bar" '
            'FROM "efg"'
            ') "sq1" ON "sq0"."foo"="sq1"."foo_two"',
            str(query),
        )

    def test_from_subquery_without_alias(self):
        subquery = Query.from_(self.table_efg).select(
            self.table_efg.base_id.as_("x"),
            self.table_efg.fizz,
            self.table_efg.buzz,
        )

        test_query = Query.from_(subquery).select(subquery.x, subquery.fizz, subquery.buzz)

        self.assertEqual(
            'SELECT "sq0"."x","sq0"."fizz","sq0"."buzz" FROM (SELECT "base_id" "x","fizz","buzz" FROM "efg") "sq0"',
            str(test_query),
        )

    def test_join_query_with_alias(self):
        subquery = (
            Query.from_(self.table_efg)
            .select(
                self.table_efg.base_id.as_("x"),
                self.table_efg.fizz,
                self.table_efg.buzz,
            )
            .as_("subq")
        )

        test_query = Query.from_(subquery).select(subquery.x, subquery.fizz, subquery.buzz)

        self.assertEqual(
            'SELECT "subq"."x","subq"."fizz","subq"."buzz" FROM (SELECT "base_id" "x","fizz","buzz" FROM "efg") "subq"',
            str(test_query),
        )

    def test_with(self):
        sub_query = Query.from_(self.table_efg).select("fizz")
        test_query = Query.with_(sub_query, "an_alias").from_(AliasedQuery("an_alias")).select("*")

        self.assertEqual(
            'WITH an_alias AS (SELECT "fizz" FROM "efg") SELECT * FROM an_alias',
            str(test_query),
        )

    def test_join_with_with(self):
        sub_query = Query.from_(self.table_efg).select("fizz")
        test_query = (
            Query.with_(sub_query, "an_alias")
            .from_(self.table_abc)
            .join(AliasedQuery("an_alias"))
            .on(AliasedQuery("an_alias").fizz == self.table_abc.buzz)
            .select("*")
        )
        self.assertEqual(
            'WITH an_alias AS (SELECT "fizz" FROM "efg") '
            'SELECT * FROM "abc" JOIN an_alias ON "an_alias"."fizz"="abc"."buzz"',
            str(test_query),
        )

    def test_select_from_with_returning(self):
        sub_query = PostgreSQLQuery.into(self.table_abc).insert(1).returning('*')
        test_query = Query.with_(sub_query, "an_alias").from_(AliasedQuery("an_alias")).select("*")
        self.assertEqual(
            'WITH an_alias AS (INSERT INTO "abc" VALUES (1) RETURNING *) SELECT * FROM an_alias',
            str(test_query),
        )


class QuoteTests(unittest.TestCase):
    def test_extraneous_quotes(self):
        t1 = Table("table1", alias="t1")
        t2 = Table("table2", alias="t2")

        query = Query.from_(t1).join(t2).on(t1.Value.between(t2.start, t2.end)).select(t1.value)

        self.assertEqual(
            "SELECT t1.value FROM table1 t1 JOIN table2 t2 ON t1.Value BETWEEN t2.start AND t2.end",
            query.get_sql(quote_char=None),
        )



import unittest

from parameterized import parameterized

from pypika import Field
from pypika.clickhouse.search_string import (
    Like,
    Match,
    MultiMatchAny,
    MultiSearchAny,
    NotLike,
)


class TestSearchString(unittest.TestCase):
    @parameterized.expand(
        [
            (Match(Field("name"), "twheys"), "match(toString(\"name\"),'twheys')"),
            (Like(Field("name"), "twheys"), "like(toString(\"name\"),'twheys')"),
            (NotLike(Field("name"), "twheys"), "notLike(toString(\"name\"),'twheys')"),
        ],
    )
    def test_search_string(self, func, expected):
        self.assertEqual(func, expected)


class TestMultiSearch(unittest.TestCase):
    @parameterized.expand(
        [
            (
                MultiSearchAny(Field("name"), ["sarah", "connor"]),
                "multiSearchAny(toString(\"name\"),['sarah','connor'])",
            ),
            (
                MultiMatchAny(Field("name"), ["sarah", "connor"]),
                "multiMatchAny(toString(\"name\"),['sarah','connor'])",
            ),
        ],
    )
    def test_multi_search_string(self, func, expected):
        self.assertEqual(func.get_sql(), expected)



import unittest

from parameterized import parameterized

from pypika import Field
from pypika.clickhouse.dates_and_times import (
    AddDays,
    AddHours,
    AddMinutes,
    AddMonths,
    AddQuarters,
    AddSeconds,
    AddWeeks,
    AddYears,
    FormatDateTime,
    SubtractDays,
    SubtractHours,
    SubtractMinutes,
    SubtractMonths,
    SubtractQuarters,
    SubtractSeconds,
    SubtractWeeks,
    SubtractYears,
    ToYYYYMM,
)


class TestFormatDateTime(unittest.TestCase):
    @parameterized.expand(
        [
            (FormatDateTime(Field("created"), "%V"), "formatDateTime(created,'%V')"),
            (
                FormatDateTime(Field("created"), "%Y-%m-%d"),
                "formatDateTime(created,'%Y-%m-%d')",
            ),
        ],
    )
    def test_get_sql(self, func, expected):
        self.assertEqual(func.get_sql(), expected)


class TestAddSubtractDt(unittest.TestCase):
    @parameterized.expand(
        [
            (AddYears(Field("created"), 1), "addYears(created,1)"),
            (AddMonths(Field("created"), 2), "addMonths(created,2)"),
            (AddWeeks(Field("created"), 3), "addWeeks(created,3)"),
            (AddDays(Field("created"), 4), "addDays(created,4)"),
            (AddHours(Field("created"), 5), "addHours(created,5)"),
            (AddMinutes(Field("created"), 6), "addMinutes(created,6)"),
            (AddSeconds(Field("created"), 7), "addSeconds(created,7)"),
            (AddQuarters(Field("created"), 8), "addQuarters(created,8)"),
            (SubtractYears(Field("created"), 9), "subtractYears(created,9)"),
            (SubtractMonths(Field("created"), 10), "subtractMonths(created,10)"),
            (SubtractWeeks(Field("created"), 11), "subtractWeeks(created,11)"),
            (SubtractDays(Field("created"), 12), "subtractDays(created,12)"),
            (SubtractHours(Field("created"), 13), "subtractHours(created,13)"),
            (SubtractHours(Field("created"), 13), "subtractHours(created,13)"),
            (SubtractMinutes(Field("created"), 14), "subtractMinutes(created,14)"),
            (SubtractSeconds(Field("created"), 15), "subtractSeconds(created,15)"),
            (SubtractQuarters(Field("created"), 16), "subtractQuarters(created,16)"),
        ],
    )
    def test_get_sql(self, func, expected):
        self.assertEqual(func.get_sql(), expected)


class TestToYYYYMM(unittest.TestCase):
    def test_get_sql(self):
        self.assertEqual(ToYYYYMM(Field("created").get_sql()), "toYYYYMM(created)")



import unittest

from pypika import (
    Case,
    CaseException,
    DatePart,
    Query,
    Schema,
    VerticaQuery,
)
from pypika import (
    Field as F,
)
from pypika import (
    Query as Q,
)
from pypika import (
    Table as T,
)
from pypika import (
    functions as fn,
)
from pypika.enums import Dialects, SqlTypes

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class FunctionTests(unittest.TestCase):
    def test_dialect_propagation(self):
        func = fn.Function("func", ["a"], ["b"])
        self.assertEqual("func(ARRAY['a'],ARRAY['b'])", func.get_sql(dialect=Dialects.POSTGRESQL))

    def test_is_aggregate_None_for_non_aggregate_function_or_function_with_no_aggregate_functions(self):
        self.assertIsNone(fn.Coalesce('a', 0).is_aggregate)
        self.assertIsNone(fn.Coalesce(fn.NullIf('a', 0), 0).is_aggregate)

    def test_is_aggregate_True_for_aggregate_function_or_function_with_aggregate_functions(self):
        self.assertTrue(fn.Sum('a').is_aggregate)
        self.assertTrue(fn.Coalesce(fn.Avg('a'), 0).is_aggregate)
        self.assertTrue(fn.Coalesce(fn.NullIf(fn.Sum('a'), 0), 0).is_aggregate)


class SchemaTests(unittest.TestCase):
    def test_schema_no_schema_in_sql_when_none_set(self):
        func = fn.Function("my_proc", 1, 2, 3)

        self.assertEqual("my_proc(1,2,3)", func.get_sql(quote_char='"'))

    def test_schema_included_in_function_sql(self):
        a = Schema("a")
        func = fn.Function("my_proc", 1, 2, 3, schema=a)

        self.assertEqual('"a".my_proc(1,2,3)', func.get_sql(quote_char='"'))


class ArithmeticTests(unittest.TestCase):
    t = T("abc")

    def test__addition__fields(self):
        q1 = Q.from_("abc").select(F("a") + F("b"))
        q2 = Q.from_(self.t).select(self.t.a + self.t.b)

        self.assertEqual('SELECT "a"+"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"+"b" FROM "abc"', str(q2))

    def test__addition__number(self):
        q1 = Q.from_("abc").select(F("a") + 1)
        q2 = Q.from_(self.t).select(self.t.a + 1)

        self.assertEqual('SELECT "a"+1 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"+1 FROM "abc"', str(q2))

    def test__addition__decimal(self):
        q1 = Q.from_("abc").select(F("a") + 1.0)
        q2 = Q.from_(self.t).select(self.t.a + 1.0)

        self.assertEqual('SELECT "a"+1.0 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"+1.0 FROM "abc"', str(q2))

    def test__addition__right(self):
        q1 = Q.from_("abc").select(1 + F("a"))
        q2 = Q.from_(self.t).select(1 + self.t.a)

        self.assertEqual('SELECT 1+"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1+"a" FROM "abc"', str(q2))

    def test__subtraction__fields(self):
        q1 = Q.from_("abc").select(F("a") - F("b"))
        q2 = Q.from_(self.t).select(self.t.a - self.t.b)

        self.assertEqual('SELECT "a"-"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"-"b" FROM "abc"', str(q2))

    def test__subtraction__number(self):
        q1 = Q.from_("abc").select(F("a") - 1)
        q2 = Q.from_(self.t).select(self.t.a - 1)

        self.assertEqual('SELECT "a"-1 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"-1 FROM "abc"', str(q2))

    def test__subtraction__decimal(self):
        q1 = Q.from_("abc").select(F("a") - 1.0)
        q2 = Q.from_(self.t).select(self.t.a - 1.0)

        self.assertEqual('SELECT "a"-1.0 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"-1.0 FROM "abc"', str(q2))

    def test__subtraction__right(self):
        q1 = Q.from_("abc").select(1 - F("a"))
        q2 = Q.from_(self.t).select(1 - self.t.a)

        self.assertEqual('SELECT 1-"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1-"a" FROM "abc"', str(q2))

    def test__multiplication__fields(self):
        q1 = Q.from_("abc").select(F("a") * F("b"))
        q2 = Q.from_(self.t).select(self.t.a * self.t.b)

        self.assertEqual('SELECT "a"*"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"*"b" FROM "abc"', str(q2))

    def test__multiplication__number(self):
        q1 = Q.from_("abc").select(F("a") * 1)
        q2 = Q.from_(self.t).select(self.t.a * 1)

        self.assertEqual('SELECT "a"*1 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"*1 FROM "abc"', str(q2))

    def test__multiplication__decimal(self):
        q1 = Q.from_("abc").select(F("a") * 1.0)
        q2 = Q.from_(self.t).select(self.t.a * 1.0)

        self.assertEqual('SELECT "a"*1.0 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"*1.0 FROM "abc"', str(q2))

    def test__multiplication__right(self):
        q1 = Q.from_("abc").select(1 * F("a"))
        q2 = Q.from_(self.t).select(1 * self.t.a)

        self.assertEqual('SELECT 1*"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1*"a" FROM "abc"', str(q2))

    def test__division__fields(self):
        q1 = Q.from_("abc").select(F("a") / F("b"))
        q2 = Q.from_(self.t).select(self.t.a / self.t.b)

        self.assertEqual('SELECT "a"/"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/"b" FROM "abc"', str(q2))

    def test__division__number(self):
        q1 = Q.from_("abc").select(F("a") / 1)
        q2 = Q.from_(self.t).select(self.t.a / 1)

        self.assertEqual('SELECT "a"/1 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/1 FROM "abc"', str(q2))

    def test__division__decimal(self):
        q1 = Q.from_("abc").select(F("a") / 1.0)
        q2 = Q.from_(self.t).select(self.t.a / 1.0)

        self.assertEqual('SELECT "a"/1.0 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/1.0 FROM "abc"', str(q2))

    def test__division__right(self):
        q1 = Q.from_("abc").select(1 / F("a"))
        q2 = Q.from_(self.t).select(1 / self.t.a)

        self.assertEqual('SELECT 1/"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1/"a" FROM "abc"', str(q2))

    def test__leftshift__fields(self):
        q1 = Q.from_("abc").select(F("a") << F("b"))
        q2 = Q.from_(self.t).select(self.t.a << self.t.b)

        self.assertEqual('SELECT "a"<<"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"<<"b" FROM "abc"', str(q2))

    def test__leftshift__number(self):
        q1 = Q.from_("abc").select(F('a') << 2)
        q2 = Q.from_(self.t).select(self.t.a << 2)

        self.assertEqual('SELECT "a"<<2 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"<<2 FROM "abc"', str(q2))

    def test__leftshift__right(self):
        q1 = Q.from_("abc").select(1 << F("a"))
        q2 = Q.from_(self.t).select(1 << self.t.a)

        self.assertEqual('SELECT 1<<"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1<<"a" FROM "abc"', str(q2))

    def test__rightshift__fields(self):
        q1 = Q.from_("abc").select(F("a") >> F("b"))
        q2 = Q.from_(self.t).select(self.t.a >> self.t.b)

        self.assertEqual('SELECT "a">>"b" FROM "abc"', str(q1))
        self.assertEqual('SELECT "a">>"b" FROM "abc"', str(q2))

    def test__rightshift__number(self):
        q1 = Q.from_("abc").select(F('a') >> 2)
        q2 = Q.from_(self.t).select(self.t.a >> 2)

        self.assertEqual('SELECT "a">>2 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a">>2 FROM "abc"', str(q2))

    def test__rightshift__right(self):
        q1 = Q.from_("abc").select(1 >> F("a"))
        q2 = Q.from_(self.t).select(1 >> self.t.a)

        self.assertEqual('SELECT 1>>"a" FROM "abc"', str(q1))
        self.assertEqual('SELECT 1>>"a" FROM "abc"', str(q2))

    def test__complex_op(self):
        q1 = Q.from_("abc").select(2 + 1 / F("a") - 5)
        q2 = Q.from_(self.t).select(2 + 1 / self.t.a - 5)

        self.assertEqual('SELECT 2+1/"a"-5 FROM "abc"', str(q1))
        self.assertEqual('SELECT 2+1/"a"-5 FROM "abc"', str(q2))

    def test__complex_op_add_parentheses(self):
        q1 = Q.from_("abc").select((F("a") + 1) + (F("b") - 5))
        q2 = Q.from_(self.t).select((self.t.a + 1) + (self.t.b - 5))
        q3 = Q.from_("abc").select((F("a") / 1) + (F("b") / 5))
        q4 = Q.from_(self.t).select((self.t.a / 1) + (self.t.b / 5))

        self.assertEqual('SELECT "a"+1+"b"-5 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"+1+"b"-5 FROM "abc"', str(q2))
        self.assertEqual('SELECT "a"/1+"b"/5 FROM "abc"', str(q3))
        self.assertEqual('SELECT "a"/1+"b"/5 FROM "abc"', str(q4))

    def test__complex_op_sub_parentheses(self):
        q1 = Q.from_("abc").select((F("a") + 1) - (F("b") - 5))
        q2 = Q.from_(self.t).select((self.t.a + 1) - (self.t.b - 5))
        q3 = Q.from_("abc").select((F("a") / 1) - (F("b") / 5))
        q4 = Q.from_(self.t).select((self.t.a / 1) - (self.t.b / 5))

        self.assertEqual('SELECT "a"+1-("b"-5) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"+1-("b"-5) FROM "abc"', str(q2))
        self.assertEqual('SELECT "a"/1-"b"/5 FROM "abc"', str(q3))
        self.assertEqual('SELECT "a"/1-"b"/5 FROM "abc"', str(q4))

    def test__complex_op_mul_parentheses(self):
        q1 = Q.from_("abc").select((F("a") + 1) * (F("b") - 5))
        q2 = Q.from_(self.t).select((self.t.a + 1) * (self.t.b - 5))
        q3 = Q.from_("abc").select((F("a") / 1) * (F("b") - 5))
        q4 = Q.from_(self.t).select((self.t.a / 1) * (self.t.b - 5))
        q5 = Q.from_("abc").select((F("a") + 1) * (F("b") / 5))
        q6 = Q.from_(self.t).select((self.t.a + 1) * (self.t.b / 5))

        self.assertEqual('SELECT ("a"+1)*("b"-5) FROM "abc"', str(q1))
        self.assertEqual('SELECT ("a"+1)*("b"-5) FROM "abc"', str(q2))
        self.assertEqual('SELECT "a"/1*("b"-5) FROM "abc"', str(q3))
        self.assertEqual('SELECT "a"/1*("b"-5) FROM "abc"', str(q4))
        self.assertEqual('SELECT ("a"+1)*"b"/5 FROM "abc"', str(q5))
        self.assertEqual('SELECT ("a"+1)*"b"/5 FROM "abc"', str(q6))

    def test__complex_op_mul_no_parentheses(self):
        q = Q.from_("abc").select(F("a") + 1 * F("b") - 5)

        self.assertEqual('SELECT "a"+1*"b"-5 FROM "abc"', str(q))

    def test__complex_op_div_parentheses(self):
        q1 = Q.from_("abc").select((F("a") + 1) / (F("b") - 5))
        q2 = Q.from_(self.t).select((self.t.a + 1) / (self.t.b - 5))
        q3 = Q.from_("abc").select((F("a") / 1) / (F("b") - 5))
        q4 = Q.from_(self.t).select((self.t.a / 1) / (self.t.b - 5))
        q5 = Q.from_("abc").select((F("a") + 1) / (F("b") * 5))
        q6 = Q.from_(self.t).select((self.t.a + 1) / (self.t.b * 5))

        self.assertEqual('SELECT ("a"+1)/("b"-5) FROM "abc"', str(q1))
        self.assertEqual('SELECT ("a"+1)/("b"-5) FROM "abc"', str(q2))
        self.assertEqual('SELECT "a"/1/("b"-5) FROM "abc"', str(q3))
        self.assertEqual('SELECT "a"/1/("b"-5) FROM "abc"', str(q4))
        self.assertEqual('SELECT ("a"+1)/("b"*5) FROM "abc"', str(q5))
        self.assertEqual('SELECT ("a"+1)/("b"*5) FROM "abc"', str(q6))

    def test__complex_op_div_no_parentheses(self):
        q = Q.from_("abc").select(F("a") + 1 / F("b") - 5)

        self.assertEqual('SELECT "a"+1/"b"-5 FROM "abc"', str(q))

    def test__complex_op_exponent_parentheses(self):
        q1 = Q.from_("abc").select(F("a") / (F("b") ** 2))
        q2 = Q.from_(self.t).select(self.t.a / (self.t.b**2))
        q3 = Q.from_("abc").select(F("a") ** (F("b") / 2))
        q4 = Q.from_(self.t).select(self.t.a ** (self.t.b / 2))
        q5 = Q.from_("abc").select((F("a") ** F("b")) ** 2)
        q6 = Q.from_(self.t).select((self.t.a**self.t.b) ** 2)

        self.assertEqual('SELECT "a"/POW("b",2) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/POW("b",2) FROM "abc"', str(q2))
        self.assertEqual('SELECT POW("a","b"/2) FROM "abc"', str(q3))
        self.assertEqual('SELECT POW("a","b"/2) FROM "abc"', str(q4))
        self.assertEqual('SELECT POW(POW("a","b"),2) FROM "abc"', str(q5))
        self.assertEqual('SELECT POW(POW("a","b"),2) FROM "abc"', str(q6))

    def test__complex_op_exponent_no_parentheses(self):
        q1 = Q.from_("abc").select(F("a") ** F("b") ** 2)
        q2 = Q.from_(self.t).select(self.t.a**self.t.b**2)
        q3 = Q.from_("abc").select(F("a") / F("b") ** 2)
        q4 = Q.from_(self.t).select(self.t.a / self.t.b**2)

        self.assertEqual('SELECT POW("a",POW("b",2)) FROM "abc"', str(q1))
        self.assertEqual('SELECT POW("a",POW("b",2)) FROM "abc"', str(q2))
        self.assertEqual('SELECT "a"/POW("b",2) FROM "abc"', str(q3))
        self.assertEqual('SELECT "a"/POW("b",2) FROM "abc"', str(q4))

    def test__complex_op_function_parentheses(self):
        q1 = Q.from_("abc").select(F("a") / (fn.Sum(F("b")) / 2))
        q2 = Q.from_("abc").select(self.t.a / (fn.Sum(self.t.b) / 2))
        q3 = Q.from_("abc").select(fn.Sum(F("a") / (F("b") / 2)))
        q4 = Q.from_("abc").select(fn.Sum(self.t.a / (self.t.b / 2)))

        self.assertEqual('SELECT "a"/(SUM("b")/2) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/(SUM("b")/2) FROM "abc"', str(q2))
        self.assertEqual('SELECT SUM("a"/("b"/2)) FROM "abc"', str(q3))
        self.assertEqual('SELECT SUM("a"/("b"/2)) FROM "abc"', str(q4))

    def test__complex_op_modulus_parentheses(self):
        q1 = Q.from_("abc").select(F("a") / (F("b") % 2))
        q2 = Q.from_(self.t).select(self.t.a / (self.t.b % 2))
        q3 = Q.from_("abc").select(F("a") % (F("b") / 2))
        q4 = Q.from_(self.t).select(self.t.a % (self.t.b / 2))
        q5 = Q.from_("abc").select(F("a") % (F("b") % 2))
        q6 = Q.from_("abc").select(self.t.a % (self.t.b % 2))

        self.assertEqual('SELECT "a"/MOD("b",2) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/MOD("b",2) FROM "abc"', str(q2))
        self.assertEqual('SELECT MOD("a","b"/2) FROM "abc"', str(q3))
        self.assertEqual('SELECT MOD("a","b"/2) FROM "abc"', str(q4))
        self.assertEqual('SELECT MOD("a",MOD("b",2)) FROM "abc"', str(q5))
        self.assertEqual('SELECT MOD("a",MOD("b",2)) FROM "abc"', str(q6))

    def test__complex_op_modulus_no_parentheses(self):
        q1 = Q.from_("abc").select(F("a") % F("b") % 2)
        q2 = Q.from_(self.t).select(self.t.a % self.t.b % 2)
        q3 = Q.from_("abc").select(F("a") / F("b") % 2)
        q4 = Q.from_(self.t).select(self.t.a / self.t.b % 2)

        self.assertEqual('SELECT MOD(MOD("a","b"),2) FROM "abc"', str(q1))
        self.assertEqual('SELECT MOD(MOD("a","b"),2) FROM "abc"', str(q2))
        self.assertEqual('SELECT MOD("a"/"b",2) FROM "abc"', str(q3))
        self.assertEqual('SELECT MOD("a"/"b",2) FROM "abc"', str(q4))

    def test__complex_op_floor_parentheses(self):
        q1 = Q.from_("abc").select(F("a") / (fn.Floor(F("b")) / 2))
        q2 = Q.from_("abc").select(self.t.a / (fn.Floor(self.t.b) / 2))
        q3 = Q.from_("abc").select(fn.Floor(F("a") / (F("b") / 2)))
        q4 = Q.from_("abc").select(fn.Floor(self.t.a / (self.t.b / 2)))

        self.assertEqual('SELECT "a"/(FLOOR("b")/2) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/(FLOOR("b")/2) FROM "abc"', str(q2))
        self.assertEqual('SELECT FLOOR("a"/("b"/2)) FROM "abc"', str(q3))
        self.assertEqual('SELECT FLOOR("a"/("b"/2)) FROM "abc"', str(q4))

    def test__complex_op_nested_parentheses(self):
        q1 = Q.from_("abc").select(F("a") / (F("b") / (F("c") / 2)))
        q2 = Q.from_("abc").select(self.t.a / (self.t.b / (self.t.c / 2)))

        self.assertEqual('SELECT "a"/("b"/("c"/2)) FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/("b"/("c"/2)) FROM "abc"', str(q2))

    def test__arithmetic_equality(self):
        q1 = Q.from_("abc").select(F("a") / 2 == 2)
        q2 = Q.from_(self.t).select(self.t.a / 2 == 2)

        self.assertEqual('SELECT "a"/2=2 FROM "abc"', str(q1))
        self.assertEqual('SELECT "a"/2=2 FROM "abc"', str(q2))

    def test__arithmetic_with_function(self):
        q1 = Q.from_("abc").select(fn.Sum(F("foo")) + 1)
        q2 = Q.from_(self.t).select(fn.Sum(self.t.foo) + 1)

        self.assertEqual('SELECT SUM("foo")+1 FROM "abc"', str(q1))
        self.assertEqual('SELECT SUM("foo")+1 FROM "abc"', str(q2))

    def test__exponent__number(self):
        q1 = Q.from_("abc").select(F("a") ** 2)
        q2 = Q.from_(self.t).select(self.t.a**2)

        self.assertEqual('SELECT POW("a",2) FROM "abc"', str(q1))
        self.assertEqual('SELECT POW("a",2) FROM "abc"', str(q2))

    def test__exponent__decimal(self):
        q1 = Q.from_("abc").select(F("a") ** 0.5)
        q2 = Q.from_(self.t).select(self.t.a**0.5)

        self.assertEqual('SELECT POW("a",0.5) FROM "abc"', str(q1))
        self.assertEqual('SELECT POW("a",0.5) FROM "abc"', str(q2))

    def test__modulus__number(self):
        q1 = Q.from_("abc").select(F("a") % 2)
        q2 = Q.from_(self.t).select(self.t.a % 2)

        self.assertEqual('SELECT MOD("a",2) FROM "abc"', str(q1))
        self.assertEqual('SELECT MOD("a",2) FROM "abc"', str(q2))

    def test__floor(self):
        q1 = Q.from_("abc").select(fn.Floor(F("foo")))
        q2 = Q.from_(self.t).select(fn.Floor(self.t.foo))

        self.assertEqual('SELECT FLOOR("foo") FROM "abc"', str(q1))
        self.assertEqual('SELECT FLOOR("foo") FROM "abc"', str(q2))


class AggregationTests(unittest.TestCase):
    def test__count(self):
        q = Q.from_("abc").select(fn.Count(F("foo")))

        self.assertEqual('SELECT COUNT("foo") FROM "abc"', str(q))

    def test__count__star(self):
        q = Q.from_("abc").select(fn.Count("*"))

        self.assertEqual('SELECT COUNT(*) FROM "abc"', str(q))

    def test__sum(self):
        q = Q.from_("abc").select(fn.Sum(F("foo")))

        self.assertEqual('SELECT SUM("foo") FROM "abc"', str(q))

    def test__avg(self):
        q = Q.from_("abc").select(fn.Avg(F("foo")))

        self.assertEqual('SELECT AVG("foo") FROM "abc"', str(q))

    def test__first(self):
        q = Q.from_("abc").select(fn.First(F("foo")))

        self.assertEqual('SELECT FIRST("foo") FROM "abc"', str(q))

    def test__last(self):
        q = Q.from_("abc").select(fn.Last(F("foo")))

        self.assertEqual('SELECT LAST("foo") FROM "abc"', str(q))

    def test__min(self):
        q = Q.from_("abc").select(fn.Min(F("foo")))

        self.assertEqual('SELECT MIN("foo") FROM "abc"', str(q))

    def test__max(self):
        q = Q.from_("abc").select(fn.Max(F("foo")))

        self.assertEqual('SELECT MAX("foo") FROM "abc"', str(q))

    def test__std(self):
        q = Q.from_("abc").select(fn.Std(F("foo")))

        self.assertEqual('SELECT STD("foo") FROM "abc"', str(q))

    def test__stddev(self):
        q = Q.from_("abc").select(fn.StdDev(F("foo")))

        self.assertEqual('SELECT STDDEV("foo") FROM "abc"', str(q))

    def test__approx_percentile(self):
        q = Q.from_("abc").select(fn.ApproximatePercentile(F("foo"), 0.5))

        self.assertEqual(
            'SELECT APPROXIMATE_PERCENTILE("foo" USING PARAMETERS percentile=0.5) FROM "abc"',
            str(q),
        )

    def test__subquery_in_params_functions(self):
        subquery = Query.from_('table').select('id')
        func = fn.Function('func', 'id', subquery)
        self.assertEqual("func('id',(SELECT id FROM table))", func.get_sql())


class ConditionTests(unittest.TestCase):
    def test__case__raw(self):
        q = Q.from_("abc").select(Case().when(F("foo") == 1, "a"))

        self.assertEqual('SELECT CASE WHEN "foo"=1 THEN \'a\' END FROM "abc"', str(q))

    def test__case__else(self):
        q = Q.from_("abc").select(Case().when(F("foo") == 1, "a").else_("b"))

        self.assertEqual("SELECT CASE WHEN \"foo\"=1 THEN 'a' ELSE 'b' END FROM \"abc\"", str(q))

    def test__case__field(self):
        q = Q.from_("abc").select(Case().when(F("foo") == 1, F("bar")).else_(F("buz")))

        self.assertEqual('SELECT CASE WHEN "foo"=1 THEN "bar" ELSE "buz" END FROM "abc"', str(q))

    def test__case__multi(self):
        q = Q.from_("abc").select(Case().when(F("foo") > 0, F("fiz")).when(F("bar") <= 0, F("buz")).else_(1))

        self.assertEqual(
            'SELECT CASE WHEN "foo">0 THEN "fiz" WHEN "bar"<=0 THEN "buz" ELSE 1 END FROM "abc"',
            str(q),
        )

    def test__case__no_cases(self):
        with self.assertRaises(CaseException):
            q = Q.from_("abc").select(Case())

            str(q)


class StringTests(unittest.TestCase):
    t = T("abc")

    def test__ascii__str(self):
        q = Q.select(fn.Ascii("2"))

        self.assertEqual("SELECT ASCII('2')", str(q))

    def test__ascii__int(self):
        q = Q.select(fn.Ascii(2))

        self.assertEqual("SELECT ASCII(2)", str(q))

    def test__ascii__field(self):
        q = Q.from_(self.t).select(fn.Ascii(self.t.foo))

        self.assertEqual('SELECT ASCII("foo") FROM "abc"', str(q))

    def test__bin__str(self):
        q = Q.select(fn.Bin("2"))

        self.assertEqual("SELECT BIN('2')", str(q))

    def test__bin__int(self):
        q = Q.select(fn.Bin(2))

        self.assertEqual("SELECT BIN(2)", str(q))

    def test__bin__field(self):
        q = Q.from_(self.t).select(fn.Bin(self.t.foo))

        self.assertEqual('SELECT BIN("foo") FROM "abc"', str(q))

    def test__concat__str(self):
        q = Q.select(fn.Concat("p", "y", "q", "b"))

        self.assertEqual("SELECT CONCAT('p','y','q','b')", str(q))

    def test__concat__field(self):
        q = Q.from_(self.t).select(fn.Concat(self.t.foo, self.t.bar))

        self.assertEqual('SELECT CONCAT("foo","bar") FROM "abc"', str(q))

    def test__insert__str(self):
        q = Q.select(fn.Insert("Quadratic", 3, 4, "What"))

        self.assertEqual("SELECT INSERT('Quadratic',3,4,'What')", str(q))

    def test__insert__field(self):
        q = Q.from_(self.t).select(fn.Insert(self.t.foo, 3, 4, self.t.bar))

        self.assertEqual('SELECT INSERT("foo",3,4,"bar") FROM "abc"', str(q))

    def test__lower__str(self):
        q = Q.select(fn.Lower("ABC"))

        self.assertEqual("SELECT LOWER('ABC')", str(q))

    def test__lower__field(self):
        q = Q.from_(self.t).select(fn.Lower(self.t.foo))

        self.assertEqual('SELECT LOWER("foo") FROM "abc"', str(q))

    def test__length__str(self):
        q = Q.select(fn.Length("ABC"))

        self.assertEqual("SELECT LENGTH('ABC')", str(q))

    def test__length__field(self):
        q = Q.from_(self.t).select(fn.Length(self.t.foo))

        self.assertEqual('SELECT LENGTH("foo") FROM "abc"', str(q))

    def test__substring(self):
        q = Q.from_(self.t).select(fn.Substring(self.t.foo, 2, 6))

        self.assertEqual('SELECT SUBSTRING("foo",2,6) FROM "abc"', str(q))

    def test__replace__str(self):
        q = Q.select(fn.Replace("ABC_DEF", "_", " "))
        self.assertEqual("SELECT REPLACE('ABC_DEF','_',' ')", str(q))

    def test__replace__field(self):
        q = Q.from_(self.t).select(fn.Replace(self.t.foobar, self.t.foo, self.t.bar))
        self.assertEqual('SELECT REPLACE("foobar","foo","bar") FROM "abc"', str(q))


class SplitPartFunctionTests(unittest.TestCase):
    t = T("abc")

    def test__split_part(self):
        q = VerticaQuery.from_(self.t).select(fn.SplitPart(self.t.foo, "|", 3))

        self.assertEqual('SELECT SPLIT_PART("foo",\'|\',3) FROM "abc"', str(q))


class RegexpLikeFunctionTests(unittest.TestCase):
    t = T("abc")

    def test__regexp_like(self):
        q = VerticaQuery.from_(self.t).select(fn.RegexpLike(self.t.foo, "^a", "x"))

        self.assertEqual("SELECT REGEXP_LIKE(\"foo\",'^a','x') FROM \"abc\"", str(q))


class CastTests(unittest.TestCase):
    t = T("abc")

    def test__cast__as(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.UNSIGNED))

        self.assertEqual('SELECT CAST("foo" AS UNSIGNED) FROM "abc"', str(q))

    def test__cast__signed(self):
        q1 = Q.from_(self.t).select(fn.Signed(self.t.foo))
        q2 = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.SIGNED))

        self.assertEqual('SELECT CAST("foo" AS SIGNED) FROM "abc"', str(q1))
        self.assertEqual('SELECT CAST("foo" AS SIGNED) FROM "abc"', str(q2))

    def test__cast__unsigned(self):
        q1 = Q.from_(self.t).select(fn.Unsigned(self.t.foo))
        q2 = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.UNSIGNED))

        self.assertEqual('SELECT CAST("foo" AS UNSIGNED) FROM "abc"', str(q1))
        self.assertEqual('SELECT CAST("foo" AS UNSIGNED) FROM "abc"', str(q2))

    def test__cast__date(self):
        q1 = Q.from_(self.t).select(fn.Date(self.t.foo))
        q2 = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.DATE))

        self.assertEqual('SELECT DATE("foo") FROM "abc"', str(q1))
        self.assertEqual('SELECT CAST("foo" AS DATE) FROM "abc"', str(q2))

    def test__cast__timestamp(self):
        q1 = Q.from_(self.t).select(fn.Timestamp(self.t.foo))
        q2 = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.TIMESTAMP))

        self.assertEqual('SELECT TIMESTAMP("foo") FROM "abc"', str(q1))
        self.assertEqual('SELECT CAST("foo" AS TIMESTAMP) FROM "abc"', str(q2))

    def test__cast__char(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.CHAR))

        self.assertEqual('SELECT CAST("foo" AS CHAR) FROM "abc"', str(q))

    def test__cast__char_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.VARCHAR(24)))

        self.assertEqual('SELECT CAST("foo" AS VARCHAR(24)) FROM "abc"', str(q))

    def test__cast__varchar(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.VARCHAR))

        self.assertEqual('SELECT CAST("foo" AS VARCHAR) FROM "abc"', str(q))

    def test__cast__varchar_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.VARCHAR(24)))

        self.assertEqual('SELECT CAST("foo" AS VARCHAR(24)) FROM "abc"', str(q))

    def test__cast__long_varchar(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.LONG_VARCHAR))

        self.assertEqual('SELECT CAST("foo" AS LONG VARCHAR) FROM "abc"', str(q))

    def test__cast__long_varchar_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.LONG_VARCHAR(24)))

        self.assertEqual('SELECT CAST("foo" AS LONG VARCHAR(24)) FROM "abc"', str(q))

    def test__cast__binary(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.BINARY))

        self.assertEqual('SELECT CAST("foo" AS BINARY) FROM "abc"', str(q))

    def test__cast__binary_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.BINARY(24)))

        self.assertEqual('SELECT CAST("foo" AS BINARY(24)) FROM "abc"', str(q))

    def test__cast__varbinary(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.VARBINARY))

        self.assertEqual('SELECT CAST("foo" AS VARBINARY) FROM "abc"', str(q))

    def test__cast__varbinary_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.VARBINARY(24)))

        self.assertEqual('SELECT CAST("foo" AS VARBINARY(24)) FROM "abc"', str(q))

    def test__cast__long_varbinary(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.LONG_VARBINARY))

        self.assertEqual('SELECT CAST("foo" AS LONG VARBINARY) FROM "abc"', str(q))

    def test__cast__long_varbinary_with_arg(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.LONG_VARBINARY(24)))

        self.assertEqual('SELECT CAST("foo" AS LONG VARBINARY(24)) FROM "abc"', str(q))

    def test__cast__boolean(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.BOOLEAN))

        self.assertEqual('SELECT CAST("foo" AS BOOLEAN) FROM "abc"', str(q))

    def test__cast__integer(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.INTEGER))

        self.assertEqual('SELECT CAST("foo" AS INTEGER) FROM "abc"', str(q))

    def test__cast__float(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.FLOAT))

        self.assertEqual('SELECT CAST("foo" AS FLOAT) FROM "abc"', str(q))

    def test__cast__numeric(self):
        q = Q.from_(self.t).select(fn.Cast(self.t.foo, SqlTypes.NUMERIC))

        self.assertEqual('SELECT CAST("foo" AS NUMERIC) FROM "abc"', str(q))

    def test__tochar__(self):
        q = Q.from_(self.t).select(fn.ToChar(self.t.foo, "SomeFormat"))

        self.assertEqual('SELECT TO_CHAR("foo",\'SomeFormat\') FROM "abc"', str(q))


class DateFunctionsTests(unittest.TestCase):
    dt = F("dt")
    t = T("abc")
    t2 = T("efg")

    def _test_extract_datepart(self, date_part):
        q = Q.from_(self.t).select(fn.Extract(date_part, self.t.foo))

        value = getattr(date_part, 'value', date_part)
        self.assertEqual('SELECT EXTRACT(%s FROM "foo") FROM "abc"' % value, str(q))

    def test_extract_microsecond(self):
        self._test_extract_datepart(DatePart.microsecond)
        self._test_extract_datepart(DatePart.microsecond.value)

    def test_extract_second(self):
        self._test_extract_datepart(DatePart.second)
        self._test_extract_datepart(DatePart.second.value)

    def test_extract_minute(self):
        self._test_extract_datepart(DatePart.minute)
        self._test_extract_datepart(DatePart.minute.value)

    def test_extract_hour(self):
        self._test_extract_datepart(DatePart.hour)
        self._test_extract_datepart(DatePart.hour.value)

    def test_extract_day(self):
        self._test_extract_datepart(DatePart.day)
        self._test_extract_datepart(DatePart.day.value)

    def test_extract_week(self):
        self._test_extract_datepart(DatePart.week)
        self._test_extract_datepart(DatePart.week.value)

    def test_extract_month(self):
        self._test_extract_datepart(DatePart.month)
        self._test_extract_datepart(DatePart.month.value)

    def test_extract_quarter(self):
        self._test_extract_datepart(DatePart.quarter)
        self._test_extract_datepart(DatePart.quarter.value)

    def test_extract_year(self):
        self._test_extract_datepart(DatePart.year)
        self._test_extract_datepart(DatePart.year.value)

    def test_extract_join(self):
        q = Q.from_(self.t).join(self.t2).on(self.t.id == self.t2.t_id).select(fn.Extract(DatePart.year, self.t.foo))

        self.assertEqual(
            'SELECT EXTRACT(YEAR FROM "abc"."foo") FROM "abc" JOIN "efg" ON "abc"."id"="efg"."t_id"',
            str(q),
        )

    def test_timestampadd(self):
        a = fn.TimestampAdd("year", 1, "2017-10-01")
        self.assertEqual(str(a), "TIMESTAMPADD(year,1,'2017-10-01')")

    def test_time_diff(self):
        a = fn.TimeDiff("18:00:00", "10:00:00")
        self.assertEqual(str(a), "TIMEDIFF('18:00:00','10:00:00')")

    def test_date_add(self):
        a = fn.DateAdd("year", 1, "2017-10-01")
        self.assertEqual(str(a), "DATE_ADD(year,1,'2017-10-01')")

    def test_now(self):
        query = Query.select(fn.Now())

        self.assertEqual("SELECT NOW()", str(query))

    def test_utc_timestamp(self):
        query = Query.select(fn.UtcTimestamp())

        self.assertEqual("SELECT UTC_TIMESTAMP()", str(query))

    def test_current_date(self):
        query = Query.select(fn.CurDate())

        self.assertEqual("SELECT CURRENT_DATE()", str(query))

    def test_current_time(self):
        query = Query.select(fn.CurTime())

        self.assertEqual("SELECT CURRENT_TIME()", str(query))

    def test_current_timestamp(self):
        query = Query.select(fn.CurTimestamp())

        self.assertEqual("SELECT CURRENT_TIMESTAMP", str(query))

    def test_current_timestamp_with_alias(self):
        query = Query.select(fn.CurTimestamp("ts"))

        self.assertEqual('SELECT CURRENT_TIMESTAMP "ts"', str(query))

    def test_to_date(self):
        q1 = fn.ToDate("2019-06-21", "yyyy-mm-dd")
        q2 = Query.from_(self.t).select(fn.ToDate("2019-06-21", "yyyy-mm-dd"))
        q3 = Query.from_(self.t).select(fn.ToDate(F("foo"), "yyyy-mm-dd"))

        self.assertEqual(str(q1), "TO_DATE('2019-06-21','yyyy-mm-dd')")
        self.assertEqual(str(q2), "SELECT TO_DATE('2019-06-21','yyyy-mm-dd') FROM \"abc\"")
        self.assertEqual(str(q3), 'SELECT TO_DATE("foo",\'yyyy-mm-dd\') FROM "abc"')


class NullFunctionsTests(unittest.TestCase):
    def test_isnull(self):
        q = Q.from_("abc").select(fn.IsNull(F("foo")))

        self.assertEqual('SELECT ISNULL("foo") FROM "abc"', str(q))

    def test_coalesce(self):
        q = Q.from_("abc").select(fn.Coalesce(F("foo"), 0))

        self.assertEqual('SELECT COALESCE("foo",0) FROM "abc"', str(q))

    def test_nullif(self):
        q = Q.from_("abc").select(fn.NullIf(F("foo"), 0))

        self.assertEqual('SELECT NULLIF("foo",0) FROM "abc"', str(q))

    def test_nvl(self):
        q = Q.from_("abc").select(fn.NVL(F("foo"), 0))

        self.assertEqual('SELECT NVL("foo",0) FROM "abc"', str(q))



import unittest

from pypika import (
    Columns,
    Database,
    Query,
    Tables,
)


class DropTableTests(unittest.TestCase):
    database_xyz = Database("mydb")
    new_table, existing_table = Tables("abc", "efg")
    foo, bar = Columns(("a", "INT"), ("b", "VARCHAR(100)"))

    def test_drop_database(self):
        q1 = Query.drop_database(self.database_xyz)
        q2 = Query.drop_database(self.database_xyz).if_exists()

        self.assertEqual('DROP DATABASE "mydb"', str(q1))
        self.assertEqual('DROP DATABASE IF EXISTS "mydb"', str(q2))

    def test_drop_table(self):
        q1 = Query.drop_table(self.new_table)
        q2 = Query.drop_table(self.new_table).if_exists()

        self.assertEqual('DROP TABLE "abc"', str(q1))
        self.assertEqual('DROP TABLE IF EXISTS "abc"', str(q2))

    def test_drop_user(self):
        q1 = Query.drop_user("myuser")
        q2 = Query.drop_user("myuser").if_exists()

        self.assertEqual('DROP USER "myuser"', str(q1))
        self.assertEqual('DROP USER IF EXISTS "myuser"', str(q2))

    def test_drop_view(self):
        q1 = Query.drop_view("myview")
        q2 = Query.drop_view("myview").if_exists()

        self.assertEqual('DROP VIEW "myview"', str(q1))
        self.assertEqual('DROP VIEW IF EXISTS "myview"', str(q2))

    def test_drop_index(self):
        q1 = Query.drop_index("myindex")
        q2 = Query.drop_index("myindex").if_exists()

        self.assertEqual('DROP INDEX "myindex"', str(q1))
        self.assertEqual('DROP INDEX IF EXISTS "myindex"', str(q2))



from pypika import CustomFunction

_add_subtract_args = ["name", "interval"]

ToYYYYMM = CustomFunction("toYYYYMM")
AddYears = CustomFunction("addYears", _add_subtract_args)
AddMonths = CustomFunction("addMonths", _add_subtract_args)
AddWeeks = CustomFunction("addWeeks", _add_subtract_args)
AddDays = CustomFunction("addDays", _add_subtract_args)
AddHours = CustomFunction("addHours", _add_subtract_args)
AddMinutes = CustomFunction("addMinutes", _add_subtract_args)
AddSeconds = CustomFunction("addSeconds", _add_subtract_args)
AddQuarters = CustomFunction("addQuarters", _add_subtract_args)
SubtractYears = CustomFunction("subtractYears", _add_subtract_args)
SubtractMonths = CustomFunction("subtractMonths", _add_subtract_args)
SubtractWeeks = CustomFunction("subtractWeeks", _add_subtract_args)
SubtractDays = CustomFunction("subtractDays", _add_subtract_args)
SubtractHours = CustomFunction("subtractHours", _add_subtract_args)
SubtractMinutes = CustomFunction("subtractMinutes", _add_subtract_args)
SubtractSeconds = CustomFunction("subtractSeconds", _add_subtract_args)
SubtractQuarters = CustomFunction("subtractQuarters", _add_subtract_args)
FormatDateTime = CustomFunction("formatDateTime", ["name", "dt_format"])



import unittest

from pypika import OracleQuery, Table
from pypika.analytics import Count


class SelectTests(unittest.TestCase):
    def test_groupby_alias_False_does_not_group_by_alias_with_standard_query(self):
        t = Table('table1')
        col = t.abc.as_('a')
        q = OracleQuery.from_(t).select(col, Count('*')).groupby(col)

        self.assertEqual('SELECT abc a,COUNT(\'*\') FROM table1 GROUP BY abc', str(q))

    def test_groupby_alias_False_does_not_group_by_alias_when_subqueries_are_present(self):
        t = Table('table1')
        subquery = OracleQuery.from_(t).select(t.abc)
        col = subquery.abc.as_('a')
        q = OracleQuery.from_(subquery).select(col, Count('*')).groupby(col)

        self.assertEqual('SELECT sq0.abc a,COUNT(\'*\') FROM (SELECT abc FROM table1) sq0 GROUP BY sq0.abc', str(q))

    def test_limit_query(self):
        t = Table('table1')
        limit = 5
        q = OracleQuery.from_(t).select(t.test).limit(limit)

        self.assertEqual(f'SELECT test FROM table1 FETCH NEXT {limit} ROWS ONLY', str(q))

    def test_offset_query(self):
        t = Table('table1')
        offset = 5
        q = OracleQuery.from_(t).select(t.test).offset(offset)

        self.assertEqual(f'SELECT test FROM table1 OFFSET {offset} ROWS', str(q))

    def test_limit_offset_query(self):
        t = Table('table1')
        limit = 5
        offset = 5
        q = OracleQuery.from_(t).select(t.test).limit(limit).offset(offset)

        self.assertEqual(f'SELECT test FROM table1 OFFSET {offset} ROWS FETCH NEXT {limit} ROWS ONLY', str(q))

    def test_fetch_next_method_deprecated(self):
        with self.assertWarns(DeprecationWarning):
            t = Table('table1')
            limit = 5
            q = OracleQuery.from_(t).select(t.test).fetch_next(limit)

            self.assertEqual(f'SELECT test FROM table1 FETCH NEXT {limit} ROWS ONLY', str(q))
